@@ -16,7 +16,7 @@
 //! Execution engine for EndBASIC programs.
 
 use crate::ast::{ArgSep, Expr, Statement, Value, VarRef, VarType};
-use crate::eval::{self, BuiltinFunction, CallableMetadata, CallableMetadataBuilder, Vars};
+use crate::eval::{self, CallableMetadata, Function, Vars};
 use crate::parser::{self, Parser};
 use async_trait::async_trait;
 use std::collections::HashMap;
@@ -71,7 +71,7 @@ pub fn new_usage_error<T, S: Into<String>>(message: S) -> Result<T> {
 /// Idiomatically, these objects need to provide a `new()` method that returns an `Rc<Callable>`, as
 /// that's the type used throughout the execution engine.
 #[async_trait(?Send)]
-pub trait BuiltinCommand {
+pub trait Command {
     /// Returns the metadata for this command.
     ///
     /// The return value takes the form of a reference to force the callable to store the metadata
@@ -91,82 +91,77 @@ pub trait BuiltinCommand {
     async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> Result<()>;
 }
 
-/// Builder pattern for a new machine.
-///
-/// The `default` constructor creates a new builder for a machine with no builtin commands and no
-/// builtin functions.
-#[derive(Default)]
-pub struct MachineBuilder {
-    commands: HashMap<&'static str, Rc<dyn BuiltinCommand>>,
-    functions: HashMap<&'static str, Rc<dyn BuiltinFunction>>,
+/// Describes how the machine stopped execution while it was running a script via `exec()`.
+#[derive(Clone, Debug, PartialEq)]
+#[must_use]
+pub enum StopReason {
+    /// Execution terminates because the machine reached the end of the input.
+    Eof,
+
+    /// Execution terminated because the machine was asked to terminate with `exit()`.
+    Exited(u8),
 }
 
-impl MachineBuilder {
-    /// Registers the given builtin command, which must not yet be registered.
-    pub fn add_command(mut self, command: Rc<dyn BuiltinCommand>) -> Self {
-        let metadata = command.metadata();
-        assert!(
-            self.commands.get(&metadata.name()).is_none(),
-            "Command with the same name already registered"
-        );
-        self.commands.insert(metadata.name(), command);
-        self
-    }
-
-    /// Registers the given builtin function, which must not yet be registered.
-    pub fn add_function(mut self, function: Rc<dyn BuiltinFunction>) -> Self {
-        let metadata = function.metadata();
-        assert!(
-            self.functions.get(&metadata.name()).is_none(),
-            "Function with the same name already registered"
-        );
-        self.functions.insert(metadata.name(), function);
-        self
-    }
-
-    /// Registers the given builtin commands.  See `add_command` for more details.
-    pub fn add_commands(mut self, commands: Vec<Rc<dyn BuiltinCommand>>) -> Self {
-        for command in commands {
-            self = self.add_command(command);
+impl StopReason {
+    /// Converts the stop reason into a process exit code.
+    pub fn as_exit_code(&self) -> i32 {
+        match self {
+            StopReason::Eof => 0,
+            StopReason::Exited(i) => *i as i32,
         }
-        self
-    }
-
-    /// Registers the given builtin functions.  See `add_function` for more details.
-    pub fn add_functions(mut self, functions: Vec<Rc<dyn BuiltinFunction>>) -> Self {
-        for function in functions {
-            self = self.add_function(function);
-        }
-        self
-    }
-
-    /// Creates a new machine with the current configuration.
-    pub fn build(self) -> Machine {
-        Machine { commands: self.commands, functions: self.functions, vars: Vars::default() }
     }
 }
 
 /// Executes an EndBASIC program and tracks its state.
 #[derive(Default)]
 pub struct Machine {
-    commands: HashMap<&'static str, Rc<dyn BuiltinCommand>>,
-    functions: HashMap<&'static str, Rc<dyn BuiltinFunction>>,
+    commands: HashMap<&'static str, Rc<dyn Command>>,
+    functions: HashMap<&'static str, Rc<dyn Function>>,
     vars: Vars,
+    stop_reason: Option<StopReason>,
 }
 
 impl Machine {
+    /// Registers the given builtin command, which must not yet be registered.
+    pub fn add_command(&mut self, command: Rc<dyn Command>) {
+        let metadata = command.metadata();
+        assert!(
+            self.commands.get(&metadata.name()).is_none(),
+            "Command with the same name already registered"
+        );
+        self.commands.insert(metadata.name(), command);
+    }
+
+    /// Registers the given builtin function, which must not yet be registered.
+    pub fn add_function(&mut self, function: Rc<dyn Function>) {
+        let metadata = function.metadata();
+        assert!(
+            self.functions.get(&metadata.name()).is_none(),
+            "Function with the same name already registered"
+        );
+        self.functions.insert(metadata.name(), function);
+    }
+
     /// Resets the state of the machine by clearing all variable.
     pub fn clear(&mut self) {
         self.vars.clear()
     }
 
+    /// Tells the machine to stop execution at the next statement boundary.
+    ///
+    /// The `exec()` call that's stopped by this invocation will return the `code` given to this
+    /// call.
+    pub fn exit(&mut self, code: u8) {
+        self.stop_reason = Some(StopReason::Exited(code));
+    }
+
     /// Obtains immutable access to the builtin commands provided by this machine.
-    pub fn get_commands(&self) -> &HashMap<&'static str, Rc<dyn BuiltinCommand>> {
+    pub fn get_commands(&self) -> &HashMap<&'static str, Rc<dyn Command>> {
         &self.commands
     }
 
     /// Obtains immutable access to the builtin functions provided by this machine.
-    pub fn get_functions(&self) -> &HashMap<&'static str, Rc<dyn BuiltinFunction>> {
+    pub fn get_functions(&self) -> &HashMap<&'static str, Rc<dyn Function>> {
         &self.functions
     }
 
@@ -285,6 +280,10 @@ impl Machine {
 
     /// Executes a single statement.
     async fn exec_one<'a>(&'a mut self, stmt: &'a Statement) -> Result<()> {
+        if self.stop_reason.is_some() {
+            return Ok(());
+        }
+
         match stmt {
             Statement::Assignment(vref, expr) => self.assign(vref, expr)?,
             Statement::BuiltinCall(name, args) => {
@@ -322,53 +321,63 @@ impl Machine {
     ///
     /// Note that this does not consume `self`.  As a result, it is possible to execute multiple
     /// different programs on the same machine, all sharing state.
-    pub async fn exec(&mut self, input: &mut dyn io::Read) -> Result<()> {
+    pub async fn exec(&mut self, input: &mut dyn io::Read) -> Result<StopReason> {
+        debug_assert!(self.stop_reason.is_none());
         let mut parser = Parser::from(input);
-        while let Some(stmt) = parser.parse()? {
-            self.exec_one(&stmt).await?;
+        while self.stop_reason.is_none() {
+            match parser.parse()? {
+                Some(stmt) => self.exec_one(&stmt).await?,
+                None => break,
+            }
         }
-        Ok(())
-    }
-}
-
-/// The `CLEAR` command.
-pub struct ClearCommand {
-    metadata: CallableMetadata,
-}
-
-impl ClearCommand {
-    /// Creates a new `CLEAR` command that resets the state of the machine.
-    pub fn new() -> Rc<Self> {
-        Rc::from(Self {
-            metadata: CallableMetadataBuilder::new("CLEAR", VarType::Void)
-                .with_syntax("")
-                .with_category("Interpreter manipulation")
-                .with_description("Clears all variables to restore initial state.")
-                .build(),
-        })
-    }
-}
-
-#[async_trait(?Send)]
-impl BuiltinCommand for ClearCommand {
-    fn metadata(&self) -> &CallableMetadata {
-        &self.metadata
-    }
-
-    async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> Result<()> {
-        if !args.is_empty() {
-            return new_usage_error("CLEAR takes no arguments");
-        }
-        machine.clear();
-        Ok(())
+        Ok(self.stop_reason.take().unwrap_or(StopReason::Eof))
     }
 }
 
 #[cfg(test)]
 pub(crate) mod testutils {
     use super::*;
+    use crate::eval::CallableMetadataBuilder;
     use std::cell::RefCell;
 
+    /// Simplified version of `EXIT` to test the machine's `exit()` method.
+    pub struct ExitCommand {
+        metadata: CallableMetadata,
+    }
+
+    impl ExitCommand {
+        /// Creates a new command that terminates execution once called.
+        pub fn new() -> Rc<Self> {
+            Rc::from(Self {
+                metadata: CallableMetadataBuilder::new("EXIT", VarType::Void).test_build(),
+            })
+        }
+    }
+
+    #[async_trait(?Send)]
+    impl Command for ExitCommand {
+        fn metadata(&self) -> &CallableMetadata {
+            &self.metadata
+        }
+
+        async fn exec(&self, args: &[(Option<Expr>, ArgSep)], machine: &mut Machine) -> Result<()> {
+            let arg = match args {
+                [(Some(expr), ArgSep::End)] => {
+                    match expr.eval(machine.get_vars(), machine.get_functions())? {
+                        Value::Integer(n) => {
+                            assert!((0..128).contains(&n), "Exit code out of range");
+                            n as u8
+                        }
+                        _ => panic!("Exit code must be a positive integer"),
+                    }
+                }
+                _ => panic!("EXIT takes one argument"),
+            };
+            machine.exit(arg);
+            Ok(())
+        }
+    }
+
     /// Simplified version of `INPUT` to feed input values based on some golden `data`.
     ///
     /// Every time this command is invoked, it yields the next value from the `data` iterator and
@@ -391,7 +400,7 @@ pub(crate) mod testutils {
     }
 
     #[async_trait(?Send)]
-    impl BuiltinCommand for InCommand {
+    impl Command for InCommand {
         fn metadata(&self) -> &CallableMetadata {
             &self.metadata
         }
@@ -436,7 +445,7 @@ pub(crate) mod testutils {
     }
 
     #[async_trait(?Send)]
-    impl BuiltinCommand for OutCommand {
+    impl Command for OutCommand {
         fn metadata(&self) -> &CallableMetadata {
             &self.metadata
         }
@@ -471,7 +480,10 @@ mod tests {
     #[test]
     fn test_clear() {
         let mut machine = Machine::default();
-        block_on(machine.exec(&mut b"a = TRUE: b = 1".as_ref())).expect("Execution failed");
+        assert_eq!(
+            StopReason::Eof,
+            block_on(machine.exec(&mut b"a = TRUE: b = 1".as_ref())).expect("Execution failed")
+        );
         assert!(machine.get_var_as_bool("a").is_ok());
         assert!(machine.get_var_as_int("b").is_ok());
         machine.clear();
@@ -482,7 +494,10 @@ mod tests {
     #[test]
     fn test_get_var_as_bool() {
         let mut machine = Machine::default();
-        block_on(machine.exec(&mut b"a = TRUE: b = 1".as_ref())).expect("Execution failed");
+        assert_eq!(
+            StopReason::Eof,
+            block_on(machine.exec(&mut b"a = TRUE: b = 1".as_ref())).expect("Execution failed")
+        );
         assert!(machine.get_var_as_bool("a").expect("Failed to query a"));
         assert_eq!(
             "Incompatible types in b? reference",
@@ -497,7 +512,10 @@ mod tests {
     #[test]
     fn test_get_var_as_int() {
         let mut machine = Machine::default();
-        block_on(machine.exec(&mut b"a = 1: b = \"foo\"".as_ref())).expect("Execution failed");
+        assert_eq!(
+            StopReason::Eof,
+            block_on(machine.exec(&mut b"a = 1: b = \"foo\"".as_ref())).expect("Execution failed")
+        );
         assert_eq!(1, machine.get_var_as_int("a").expect("Failed to query a"));
         assert_eq!(
             "Incompatible types in b% reference",
@@ -512,7 +530,11 @@ mod tests {
     #[test]
     fn test_get_var_as_string() {
         let mut machine = Machine::default();
-        block_on(machine.exec(&mut b"a = \"foo\": b = FALSE".as_ref())).expect("Execution failed");
+        assert_eq!(
+            StopReason::Eof,
+            block_on(machine.exec(&mut b"a = \"foo\": b = FALSE".as_ref()))
+                .expect("Execution failed")
+        );
         assert_eq!("foo", machine.get_var_as_string("a").expect("Failed to query a"));
         assert_eq!(
             "Incompatible types in b$ reference",
@@ -532,12 +554,12 @@ mod tests {
         input: &str,
         golden_in: &'static [&'static str],
         captured_out: Rc<RefCell<Vec<String>>>,
-    ) -> Result<()> {
-        let mut machine = MachineBuilder::default()
-            .add_command(InCommand::new(Box::from(RefCell::from(golden_in.iter()))))
-            .add_command(OutCommand::new(captured_out))
-            .add_function(SumFunction::new())
-            .build();
+    ) -> Result<StopReason> {
+        let mut machine = Machine::default();
+        machine.add_command(ExitCommand::new());
+        machine.add_command(InCommand::new(Box::from(RefCell::from(golden_in.iter()))));
+        machine.add_command(OutCommand::new(captured_out));
+        machine.add_function(SumFunction::new());
         block_on(machine.exec(&mut input.as_bytes()))
     }
 
@@ -551,7 +573,10 @@ mod tests {
         expected_out: &'static [&'static str],
     ) {
         let captured_out = Rc::from(RefCell::from(vec![]));
-        run(input, golden_in, captured_out.clone()).expect("Execution failed");
+        assert_eq!(
+            StopReason::Eof,
+            run(input, golden_in, captured_out.clone()).expect("Execution failed")
+        );
         assert_eq!(expected_out, captured_out.borrow().as_slice());
     }
 
@@ -606,6 +631,66 @@ mod tests {
         do_simple_error_test("a? = 3", "Incompatible types in a? assignment");
     }
 
+    #[test]
+    fn test_exit_simple() {
+        let captured_out = Rc::from(RefCell::from(vec![]));
+        assert_eq!(
+            StopReason::Exited(5),
+            run("OUT 1\nEXIT 5\nOUT 2", &[], captured_out.clone()).expect("Execution failed")
+        );
+        assert_eq!(&["1"], captured_out.borrow().as_slice());
+    }
+
+    #[test]
+    fn test_exit_zero_is_not_special() {
+        let captured_out = Rc::from(RefCell::from(vec![]));
+        assert_eq!(
+            StopReason::Exited(0),
+            run("OUT 1\nOUT 2\nEXIT 0\nOUT 3", &[], captured_out.clone())
+                .expect("Execution failed")
+        );
+        assert_eq!(&["1", "2"], captured_out.borrow().as_slice());
+    }
+
+    #[test]
+    fn test_exit_nested() {
+        let captured_out = Rc::from(RefCell::from(vec![]));
+        assert_eq!(
+            StopReason::Exited(42),
+            run(
+                "FOR a = 0 TO 10\nOUT a\nIF a = 3 THEN\nEXIT 42\nEND IF\nNEXT",
+                &[],
+                captured_out.clone()
+            )
+            .expect("Execution failed")
+        );
+        assert_eq!(&["0", "1", "2", "3"], captured_out.borrow().as_slice());
+    }
+
+    #[test]
+    fn test_exit_can_resume() {
+        let captured_out = Rc::from(RefCell::from(vec![]));
+        let mut machine = Machine::default();
+        machine.add_command(ExitCommand::new());
+        machine.add_command(OutCommand::new(captured_out.clone()));
+        machine.add_function(SumFunction::new());
+
+        assert_eq!(
+            StopReason::Exited(10),
+            block_on(machine.exec(&mut "OUT 1\nEXIT 10\nOUT 2".as_bytes()))
+                .expect("Execution failed")
+        );
+        assert_eq!(&["1"], captured_out.borrow().as_slice());
+
+        captured_out.borrow_mut().clear();
+        assert_eq!(
+            StopReason::Exited(11),
+            block_on(machine.exec(&mut "OUT 2\nEXIT 11\nOUT 3".as_bytes()))
+                .expect("Execution failed")
+        );
+        assert_eq!(&["2"], captured_out.borrow().as_slice());
+    }
+
     #[test]
     fn test_if_ok() {
         let code = r#"
@@ -828,25 +913,13 @@ mod tests {
     #[test]
     fn test_exec_shares_state() {
         let mut machine = Machine::default();
-        block_on(machine.exec(&mut b"a = 10".as_ref())).expect("Execution failed");
-        block_on(machine.exec(&mut b"b = a".as_ref())).expect("Execution failed");
-    }
-
-    #[test]
-    fn test_clear_ok() {
-        let mut machine = MachineBuilder::default().add_command(ClearCommand::new()).build();
-        block_on(machine.exec(&mut b"a = 1".as_ref())).unwrap();
-        assert!(machine.get_var_as_int("a").is_ok());
-        block_on(machine.exec(&mut b"CLEAR".as_ref())).unwrap();
-        assert!(machine.get_var_as_int("a").is_err());
-    }
-
-    #[test]
-    fn test_clear_errors() {
-        let mut machine = MachineBuilder::default().add_command(ClearCommand::new()).build();
         assert_eq!(
-            "CLEAR takes no arguments",
-            format!("{}", block_on(machine.exec(&mut b"CLEAR 123".as_ref())).unwrap_err())
+            StopReason::Eof,
+            block_on(machine.exec(&mut b"a = 10".as_ref())).expect("Execution failed")
+        );
+        assert_eq!(
+            StopReason::Eof,
+            block_on(machine.exec(&mut b"b = a".as_ref())).expect("Execution failed")
         );
     }
 }
